unit Mandelbrot;
{
  (c)2013 by Paul TOTH <tothpaul@free.fr>
  http://lookinside.free.fr/delphi.php?Multithreading
  
  based on Eric Grange article
  http://delphitools.info/2011/09/02/first-look-at-xe2-floating-point-performance/
}

interface

uses
  Windows;

const
  IMAGE_SIZE = 480;
  MAX_COLORS = 512;

  CONTROL_COLORS : array[0..4, 0..2] of Byte = (
    ($20, $00, $00),
    ($ff, $ff, $ff),
    ($A0, $00, $00),
    ($FF, $ff, $40),
    ($FF, $20, $20)
  );


var
  pixels : array[0..IMAGE_SIZE - 1, 0..IMAGE_SIZE - 1] of Cardinal;
  DIBInfo: TBitmapInfo;

  iterLimit : Integer;
  qmin, qmax, pmin, pmax : Double;
   xstep, ystep : Double;

  colors   : array[0..MAX_COLORS - 1] of Cardinal;

  offset   : Integer;
  cores    : Integer;
  tasks    : Integer;
  counters : array of Integer;

procedure Reset;
procedure Compute(AdditionalThreadCount: Integer);
procedure ComputePixel(sx, sy: Integer);

implementation

procedure ComputePixel(sx, sy: Integer);
const
   kmax = 256;
var
   x, y, r : Double;
   k : Integer;
   p, q, x0, y0 : Double;
begin
  p := pmin + xstep * sx;
  q := qmax - ystep * sy;
  k := 0;
  x0 := 0;
  y0 := 0;

  repeat
    x := x0 * x0 - y0 * y0 + p;
    y := 2 * x0 * y0 + q;
    x0 := x;
    y0 := y;
    r := x * x + y * y;
    Inc(k);
  until ((r > iterLimit) or (k >= kmax));

  if k >= kmax then
    k := 0;

  pixels[sy, sx] := colors[k];
end;

function MandelThread(id: Integer): Integer; stdcall;
var
  ofs: Integer;
begin
  ofs := InterlockedIncrement(offset);
  while ofs < IMAGE_SIZE * IMAGE_SIZE do
  begin
    ComputePixel(ofs mod IMAGE_SIZE, ofs div IMAGE_SIZE);
    Inc(counters[id]);
    ofs := InterlockedIncrement(offset);
  end;
  InterlockedDecrement(tasks);
end;

procedure Compute(AdditionalThreadCount: Integer);
var
  i : Integer;
  id: Cardinal;
begin
  xstep := (pmax - pmin) / IMAGE_SIZE;
  ystep := (qmax - qmin) / IMAGE_SIZE;

  offset := -1;
  tasks := AdditionalThreadCount;

  counters := nil; // reset
  SetLength(counters, AdditionalThreadCount + 1);

  IsMultiThread := True;
  for i := 1 to AdditionalThreadCount  do
    CreateThread(nil, 0, @MandelThread, Pointer(i), 0, id);

  MandelThread(0);

  while tasks > 0 do  // should no be too long, there's no more pixel to compute :)
    Sleep(0);
end;

procedure InitColors;
var
   i, k : Integer;
   rstep, bstep, gstep : Double;
begin
  colors[0] := RGB(0, 0, 0);

  for i := 0 to High(CONTROL_COLORS) do
  begin
    rstep := (CONTROL_COLORS[i + 1, 0] - CONTROL_COLORS[i, 0]) / 63;
    gstep := (CONTROL_COLORS[i + 1, 1] - CONTROL_COLORS[i, 1]) / 63;
    bstep := (CONTROL_COLORS[i + 1, 2] - CONTROL_COLORS[i, 2]) / 63;

    for k := 0 to 63 do
      colors[k + (i * 64) + 1] := RGB(Round(CONTROL_COLORS[i, 0] + rstep * k),
                                      Round(CONTROL_COLORS[i, 1] + gstep * k),
                                      Round(CONTROL_COLORS[i, 2] + bstep * k));
  end;

  for i := 257 to MAX_COLORS-1 do
    colors[i] := colors[i - 256];
end;

procedure Reset();
begin
  iterLimit := 100;
  qmin := -1.5;
  qmax := 1.5;
  pmin := -2.25;
  pmax := 0.75;
end;

procedure init();
var
  SI: SYSTEM_INFO;
begin
  FillChar(DIBInfo, SizeOf(DIBInfo), 0);
  DIBInfo.bmiHeader.biSize := 40;
  DIBInfo.bmiHeader.biWidth := IMAGE_SIZE;
  DIBInfo.bmiHeader.biHeight := - IMAGE_SIZE;
  DIBInfo.bmiHeader.biPlanes := 1;
  DIBInfo.bmiHeader.biBitCount := 32;

  GetSystemInfo(SI);
  cores := SI.dwNumberOfProcessors;

  InitColors;

  Reset();
end;

initialization
  init();
end.

